#version 100
precision mediump float;
uniform vec2 u_offset;
uniform vec2 u_lookat;
uniform float u_point_size;
uniform float u_scale;
uniform vec2 u_top;                      // The top-right geo location visible in the viewport.
uniform vec2 u_bottom;                   // The bottom-left geo location visible in the viewport.
uniform vec2 u_viewport_size;            // The viewport dimensions in pixels.
uniform vec2 u_center;
uniform vec2 u_texture_size;
uniform int u_scale_texture;

attribute vec2 a_vertex;                 // The vertices that need to be drawn as geo locations.
attribute vec2 a_texture_coordinates;    // The texture coordinates.

varying vec2 v_texcoord;                 // The texture coordinates, passed to the fragment shader.

#define M_PI 3.1415926535897932384626433832795
#define ONE_DEGREE_IN_INCHES 4374448.819
#define DPI 160.0

struct ProjectedData {
    vec4 position;                 // Projected geo point in OpenGL domain.
    vec4 center_position;          // Projected center position of the polygon.
    vec2 lookat;                   // Projected look at point.
    vec2 size;                     // Projected difference between top and bottom.
};

float project(float value) {
    return log(tan(M_PI / 4.0 + radians(value) / 2.0));
}

vec2 project(vec2 value) {
    return vec2(value.x, project(value.y));
}

vec4 projectFar(ProjectedData data, vec2 target) {
    vec2 point = 2.0 * (project(target) - data.lookat) / data.size;
    return vec4(point, 0.0, 1.0);
}

vec4 projectNear(ProjectedData data, vec2 target) {
    vec2 world = target - u_lookat;
    //TODO: Calculate the below in Java and feed to shader as one uniform value.
    float factor = 1.0 / u_scale;
    vec2 inches = world * factor * vec2(ONE_DEGREE_IN_INCHES * cos(radians(u_lookat.y)), ONE_DEGREE_IN_INCHES);
    vec2 pixels = inches * DPI;
    vec2 result = pixels / vec2(u_viewport_size.x / 2.0, u_viewport_size.y / 2.0);
    return vec4(result, 0.0, 1.0);
}

void getPosition(inout ProjectedData data) {
    //TODO: Since both approaches show a deviation at one end of the zoom scale, we
    //      switch projection algorithm in the middle of this scale. This technical debt has
    //      been logged as issue ITALIA-7921.
    if (u_scale >= 512000.0) {
        data.position = projectFar(data, a_vertex);
    } else {
        data.position = projectNear(data, a_vertex);
    }
}

void getCenterPosition(inout ProjectedData data) {
    data.center_position = projectNear(data, u_center);
}

void main()
{
    ProjectedData data;
    data.lookat = project(u_lookat);
    if (u_top.x < u_bottom.x) {
        // Passing the Eastern border so compensate for unexpected negative longitude.
        data.size = project(u_top + vec2(360.0, 0.0)) - project(u_bottom);
    } else {
        data.size = project(u_top) - project(u_bottom);
    }
    getPosition(data);
    gl_Position = data.position;
    gl_PointSize = u_point_size;

    if (u_scale_texture == 1) {
        v_texcoord = a_texture_coordinates;
    } else {
        getCenterPosition(data);
        /*
         * Half the viewport width divided by the texture width, will give us the unit of
         * texture coordinates to use, which takes the different aspect ratios into account.
         * Same for the heights, obviously.
         *
         * <pre>
         *   Wv = <width of the viewport in pixels>
         *   Wt = <width of the texture in pixels>
         *   PPUv = 1/2 * Wv <pixels per OGL unit for the viewport>
         *   PPUt = PPUv / Wt = (1/2 * Wv) / Wt = Wv / Wt / 2 <pixels per OGL unit for the texture>
         * </pre>
         */
        vec4 plane_s = vec4(u_viewport_size.x / u_texture_size.x / 2.0, 0.0, 0.0, 0.0);
        vec4 plane_t = vec4(0.0, u_viewport_size.y / u_texture_size.y / 2.0, 0.0, 0.0);

        // Adding half a unit to hook the center of the texture to the center of the polygon.
        v_texcoord.s = dot(-1.0 * (data.center_position - gl_Position), plane_s) + 0.5;
        v_texcoord.t = dot((data.center_position - gl_Position), plane_t) + 0.5;
    }
}
